#!/usr/bin/env python3
r"""
Used by CI to install a new cluster and run tests.

Env vars needed:
- COSMOS_API_TOKEN

Env vars optional:
- NO_TERMINATE_DROPLETS: list of droplet ids that should not be
  removed when deleting droplets.

Install requirements:

- On Alpine, install with:

    apk --no-cache add python3-dev build-base libffi-dev linux-headers \
      openssl-dev openssh-client
    pip3 install -r requirements.txt
"""

import argparse
import logging
from math import floor
import os
import sys
import greenhost_cloud
from behave.__main__ import main as behave_main
from openappstack import name, cluster, ansible

ALL_TESTS = ['behave']

# We're limiting to 50, because we use subdomains, the current longest of which
# is 7 characters (`office.`). Max CN length is 63 characters, so we have some
# wiggle room for longer subdomains in the future.
# See https://github.com/jetstack/cert-manager/issues/2794 for details.
MAX_DOMAIN_LENGTH = 50


def main():  # pylint: disable=too-many-statements,too-many-branches,too-many-locals
    """
    Parse arguments and start program. Depending on the arguments parsed,
    one of three functions is called:

    - create, responsible for setting up VPSs and their local configuration
      files
    - install, responsible for setting up Kubernetes and OpenAppStack on those
      VPSs
    - test, responsible for testing if all the setup steps worked.

    Run python3 -m openappstack --help for more information.
    """
    # Parse command line arguments
    parser = argparse.ArgumentParser(
        prog=name,
        description='Run installation script '
        'to deploy Openappstack to a given node. Note that you always need to '
        'provide a cluster name. For subcommand specific help, for example, '
        'run "python3 -m openappstack your-cluster install --help"')

    parser.add_argument(
        'cluster_name',
        metavar='CLUSTER_NAME',
        type=str,
        help='Name of the cluster you want to use openappstack with')

    subparsers = parser.add_subparsers(help='Available subcommands')

    create_parser = subparsers.add_parser('create', help='Creates a new cluster')
    create_parser.set_defaults(func=create)

    create_parser.add_argument(
        'domain',
        metavar='DOMAIN_NAME',
        help='Domain name to run OpenAppStack under')
    create_parser.add_argument(
        '--create-hostname',
        type=str,
        help='hostname of the machine. If not provided for a new machine, the '
             'cluster name is used.')

    create_parser.add_argument(
        '--docker-mirror-server',
        help=("Server name for a docker mirror"))
    create_parser.add_argument(
        '--docker-mirror-endpoint',
        help=("Full endpoint for a docker mirror"))
    create_parser.add_argument(
        '--docker-mirror-username',
        help=("Username for `docker login` to docker mirror"))
    create_parser.add_argument(
        '--docker-mirror-password',
        help=("Password for `docker login` to docker mirror"))


    group = create_parser.add_mutually_exclusive_group(required=True)

    group.add_argument(
        '--create-droplet',
        action='store_true',
        help='Create droplet automatically')
    group.add_argument(
        '--droplet-id',
        metavar='ID',
        type=int,
        help='ID of existing Greenhost VPS to use')
    group.add_argument(
        '--droplet-hostname',
        metavar='HOSTNAME',
        help='Hostname of existing Greenhost VPS to use')
    group.add_argument(
        '--ip-address',
        metavar='IP',
        help='IP address of non-greenhost VPS to use. You *must* provide '
             '--create-hostname with this argument')

    droplet_creation_group = create_parser.add_argument_group(
        'droplet creation',
        'When using --create-droplet, you need to provide:')

    droplet_creation_group.add_argument(
        '--ssh-key-id',
        metavar='ID',
        type=int,
        help='Greenhost-specific database ID of ssh key to deploy with')

    droplet_creation_group.add_argument(
        '--create-domain-records',
        action='store_true',
        help=('Creates DNS entries at Greenhost pointing the subdomain and '
              'domain to the cluster.'))

    droplet_creation_group.add_argument(
        '--subdomain',
        type=str,
        metavar='SUBDOMAIN',
        help=('Use a custom subdomain for the generated domain records. '
              'Defaults to no subdomain'))

    # Use this option to make sure that you won't run into troubles with
    # certificates. Let's Encrypt requires a CN for each certificate, and a
    # CN has a max length of 63 characters.
    droplet_creation_group.add_argument(
        '--truncate-subdomain',
        action='store_true',
        help=('Truncate subdomain so subdomain and domain are shorter than '
              '{} characters.'.format(MAX_DOMAIN_LENGTH)))

    droplet_creation_group.add_argument(
        '--acme-staging',
        action='store_true',
        help=("Use this for development clusters. Uses Let's Encrypt's "
              'staging environment'))

    droplet_creation_group.add_argument(
        '--local-flux',
        action='store_true',
        help=("Use this for development clusters. Uses an in-cluster "
              'auto-update feed'))

    droplet_creation_group.add_argument(
        '--disk-image-id',
        help=("Custom disk image ID. Use negative value for a custom template "
              "ID, use positive values for operating system IDs"))

    install_parser = subparsers.add_parser(
        'install',
        help=("Use this to run the ansible playbook that sets up your VPS to run "
              "OpenAppStack. The ansible-playbook process will run in the "
              "'{ansible_path}' directory, so do not use relative paths with these "
              "arguments.").format(ansible_path=ansible.ANSIBLE_PATH))
    install_parser.set_defaults(func=install)

    parser.add_argument(
        '--terminate-droplet',
        action='store_true',
        help=('Shutdown AND DELETE droplet identified by VPS identifier after '
              'finishing'))

    parser.add_argument(
        '-v',
        '--verbose',
        action='store_true',
        help='Be more verbose')

    # Ansible related
    install_parser.add_argument(
        '--ansible-param',
        metavar=['PARAM[=VALUE]'],
        action='append',
        nargs=1,
        help='forward ansible parameters to the ansible-playbook call. If '
          '--install-kubernetes is supplied, parameters are passed to both '
          'ansible playbooks')

    install_parser.add_argument(
        '--install-kubernetes',
        action='store_true',
        help="Installs k3s on your VPS before installing OpenAppStack")

    install_parser.add_argument(
        '--no-install-openappstack',
        action='store_true',
        help=("Skip openappstack installation. This is useful if you only "
              "want a kubernetes cluster."))

    test_parser = subparsers.add_parser(
        'test',
        help=("Write test configuration and run tests on your cluster"))

    test_parser.set_defaults(func=test)

    test_parser.add_argument(
        '--run-test',
        action='append',
        help=('Run only a specific kind of test. If not provided, all tests '
              'are run. The currently available tests are: ') \
            + ','.join(ALL_TESTS))

    test_parser.add_argument(
        '--behave-rerun-failing',
        action='store_true',
        help=('Run behave with @rerun_failing.features'))
    test_parser.add_argument(
        '--behave-tags',
        nargs='+',
        default=[],
        help=('Only run behave tests with these tags'))
    test_parser.add_argument(
        '--behave-headless',
        action='store_true',
        help=('Run behave in headless mode'))
    test_parser.add_argument(
        '--behave-ignore-certificate-errors',
        action='store_true',
        help=('Ignore certificate errors, needed if you use i.e. the Letsencrypt ACME staging CA'))
    test_parser.add_argument(
        '--behave-param',
        metavar=['PARAM[=VALUE]'],
        action='append',
        nargs=1,
        default=[],
        help=('Pass additional behave options (like "-D browser=firefox"'))

    info_parser = subparsers.add_parser(
        'info',
        help=("Show information about a cluster"))

    info_parser.set_defaults(func=info)
    info_parser.add_argument(
        '--ip-address',
        action='store_true',
        help=('Only output the IP address of the machine')
    )

    args = parser.parse_args()
    loglevel = logging.DEBUG if args.verbose else logging.INFO
    init_logging(log, loglevel)

    # Setup logging for greenhost_cloud module
    log_greenhost_cloud = logging.getLogger('greenhost_cloud')
    init_logging(log_greenhost_cloud, loglevel)

    log.debug("Parsed arguments: %s", str(args))

    clus = cluster.Cluster(args.cluster_name)

    if hasattr(args, 'func'):
        args.func(clus, args)

    if args.terminate_droplet:
        # In case none of the subparser's functions have been called, load data
        clus.load_data()
        greenhost_cloud.terminate_droplets_by_name('^{}$'.format(clus.hostname))

    if not hasattr(args, 'func') and not args.terminate_droplet:
        parser.print_help()
        sys.exit(1)

def info(clus, args):
    """
    Shows cluster information and then exits

    :param cluster.Cluster clus: cluster to show information about
    :param argparse.Namespace args: If the --ip-address argument is given, only
        prints the machine's IP address.
    """
    clus.load_data()
    clus.print_info(args)


def create(clus, args):  # pylint: disable=too-many-branches
    """
    Parses arguments for the 'create' subcommand

    :param cluster.Cluster clus: Cluster object to edit
    :param argparse.Namespace args: Command line arguments
    """
    clus = cluster.Cluster(args.cluster_name)

    if args.create_droplet:
        if not args.ssh_key_id:
            log.error('--ssh-key-id required when using --create-droplet')
            sys.exit(1)

    if args.subdomain:
        # Check if {subdomain}.{domain} is not too long to fit in CN
        if len(args.domain) > MAX_DOMAIN_LENGTH:
            log.error(('ERROR: domain argument is too long. Domain will not '
                       'fit in a CN, please make sure your subdomain + '
                       'domain + 1 do not exceed %d '
                       'characters'), MAX_DOMAIN_LENGTH)
            sys.exit(1)

        subdomain = args.subdomain
        if len(args.domain) + len(args.subdomain) + 1 > MAX_DOMAIN_LENGTH:
            if args.truncate_subdomain:
                half_required_length = floor((MAX_DOMAIN_LENGTH - len(args.domain) - 1)/2)
                subdomain = args.subdomain[:half_required_length]
                subdomain += args.subdomain[-half_required_length:]
                log.warning('Subdomain truncated to "%s"', subdomain)
            else:
                log.error(('ERROR: --subdomain argument is too long. Domain '
                           'will not fit in a CN, please make sure your '
                           'subdoman + domain + 1 do not exceed %d '
                           'characters'), MAX_DOMAIN_LENGTH)
                sys.exit(1)

        fqdn = '{subdomain}.{domain}'.format(
            subdomain=subdomain, domain=args.domain)
    else:
        fqdn = args.domain
        subdomain = None

    clus.domain = fqdn

    if args.disk_image_id:
        clus.disk_image_id = args.disk_image_id

    # Set acme_staging to False so we use Let's Encrypt's live environment
    if args.acme_staging:
        clus.acme_staging = True
    if args.create_droplet:
        clus.create_droplet(ssh_key_id=args.ssh_key_id, hostname=args.create_hostname)
        if args.verbose:
            greenhost_cloud.list_droplets()
        # Wait for ssh
        greenhost_cloud.wait_for_ssh(clus.ip_address)
    elif args.droplet_id:
        clus.set_info_by_droplet_id(args.droplet_id)
    elif args.ip_address:
        if not args.create_hostname:
            log.error('--create-hostname required when using --ip-address')
            sys.exit(2)
        clus.set_info_by_ip_and_hostname(args.ip_address, args.create_hostname)
    elif args.droplet_hostname:
        clus.set_info_by_hostname(args.droplet_hostname)

    if args.docker_mirror_server:
        clus.docker_mirror_server = args.docker_mirror_server
    if args.docker_mirror_endpoint:
        clus.docker_mirror_endpoint = args.docker_mirror_endpoint
    if args.docker_mirror_username:
        clus.docker_mirror_username = args.docker_mirror_username
    if args.docker_mirror_password:
        clus.docker_mirror_password = args.docker_mirror_password

    # Write inventory.yml and settings.yml files
    clus.write_cluster_files()

    if args.create_domain_records:
        create_domain_records(
            args.domain, clus.ip_address, subdomain=subdomain)
        if args.verbose:
            greenhost_cloud.list_domain_records(args.domain)


def install(clus, args):
    """
    Parses arguments for the 'install' subcommand

    :param cluster.Cluster clus: Cluster object to install OAS to
    :param argparse.Namespace args: Command line arguments
    """
    clus.load_data()

    if args.install_kubernetes:
        ansible.run_ansible(
            clus,
            os.path.join(ansible.ANSIBLE_PATH, 'install-kubernetes.yml'),
            args.ansible_param)

    if not args.no_install_openappstack:
        ansible.run_ansible(
            clus,
            os.path.join(ansible.ANSIBLE_PATH, 'install-openappstack.yml'),
            args.ansible_param)


def test(clus, args):
    """
    Runs behave or testinfra test. Overwrites behave_path/behave.ini!

    :param cluster.Cluster clus: Cluster object to run tests on
    :param argparse.Namespace args: Command line arguments
    """

    # At the moment we only have one type if test, but more tests will be added
    # to this in the future. If args.run_test is empty, we run all the tests
    run_test = args.run_test
    if not run_test:
        run_test = ALL_TESTS

    if 'behave' in run_test:
        behave_path = os.path.join(os.path.dirname(__file__), '..', 'test',
                                   'behave')
        # Run from the behave directory so behave automatically loads all the
        # necessary files
        os.chdir(behave_path)
        clus.load_data()
        behave_ini = os.path.join(behave_path, 'behave.ini')
        clus.write_behave_config(behave_ini)
        command = []
        if args.behave_rerun_failing:
            command.append('@rerun_failing.features')
        if args.behave_headless:
            command.append('-D headless=True')
        if args.behave_ignore_certificate_errors:
            command.append('-D ignore_certificate_errors=True')
        for tag in args.behave_tags:
            log.info(command)
            command.append('-t {tag}'.format(tag=tag))
        if args.behave_param:
            for param in args.behave_param:
                if len(param) > 1:
                    log.warning('More than 1 parameter. Ignoring the rest! '
                                'Use --behave-param several times to supply '
                                'more than 1 parameter')
                param = param[0]
                command.append(param)
        log.info('Running behave command %s', command)
        return_code = behave_main(command)

        # Remove behave.ini so we don't leave secrets hanging around.
        os.remove(behave_ini)

        sys.exit(return_code)

def create_domain_records(domain, droplet_ip, subdomain=None):
    """
    Creates 2 domain records at Greenhost. An A record at subdomain.domain,
    pointing to droplet_ip and a CNAME record pointing *.subdomain.domain to
    the first record.

    :param str domain: the domain the cluster will be hosted on.
    :param str droplet_ip: The IP address the A record will point to.
    :param subdomain: Optional subdomain to host OAS on.
    :type subdomain: str or None
    """
    subdomain_arg = subdomain
    if subdomain is None:
        subdomain_arg = "@"

    domain_record = greenhost_cloud.create_domain_record(
        domain=domain, name=subdomain_arg,
        data=droplet_ip, record_type='A', update=True, ttl=60)
    log.info('Domain record: %s', domain_record)

    subdomain_arg = '*'
    if subdomain is not None:
        subdomain_arg += '.' + subdomain

    domain_record = greenhost_cloud.create_domain_record(
        domain=domain, name=subdomain_arg,
        data=subdomain, record_type='CNAME', update=True)
    log.info('Domain record: %s', domain_record)


def init_logging(logger, loglevel):
    """
    Configure logging.

    - debug and info go to stdout
    - warning and above go to stderr

    :param logger: Logger to initialise
    :param int loglevel: Log level from 0 to 50 (use logging.WARNING,
        logging.INFO, etc.)
    """
    logger.setLevel(loglevel)
    stdout = logging.StreamHandler(sys.stdout)
    stdout.setLevel(loglevel)
    stdout.addFilter(lambda record: record.levelno <= logging.INFO)
    logger.info('Set log level to %d', loglevel)

    stderr = logging.StreamHandler()
    stderr.setLevel(logging.WARNING)

    logger.addHandler(stdout)
    logger.addHandler(stderr)


if __name__ == '__main__':
    # Setup logging for this script
    log = logging.getLogger('openappstack')  # pylint: disable=invalid-name
    main()
